<?php

declare(strict_types=1);

namespace Drupal\message_recipient_group\Plugin\MessageRecipient\RecipientsCollector;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\group\Entity\GroupInterface;
use Drupal\group\Entity\GroupRelationship;
use Drupal\group\Entity\GroupRelationshipInterface;
use Drupal\group\Entity\GroupTypeInterface;
use Drupal\group\Entity\Storage\GroupRelationshipTypeStorageInterface;
use Drupal\group\PermissionScopeInterface;
use Drupal\message\MessageInterface;
use Drupal\message\MessageTemplateInterface;
use Drupal\message_recipient\Attribute\RecipientsCollector;
use Drupal\message_recipient_entity\ContentEntityUserRecipientsCollectorBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Users by group membership recipients collector.
 */
#[RecipientsCollector(
  id: 'user_by_group_membership',
  type: 'user',
  category: 'group',
  label: new TranslatableMarkup('Users by group membership'),
  description: new TranslatableMarkup('Collects users by group membership from the message\'s associated "group/group relationship/group content" entity.'),
)]
class UsersByGroupMembership extends ContentEntityUserRecipientsCollectorBase {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * The user ids.
   *
   * @var array
   */
  protected array $userIds = [];

  /**
   * The soft exclusion id.
   *
   * @const string
   */
  const EXCLUSION_SOFT = 'soft';

  /**
   * The exclusion id.
   *
   * @const string
   */
  const EXCLUSION_STRICT = 'strict';

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->database = $container->get('database');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function collect(MessageInterface $message): array {
    $groups = $this->getGroupByMessage($message);
    $grt_storage = $this->entityTypeManager->getStorage('group_content_type');
    assert($grt_storage instanceof GroupRelationshipTypeStorageInterface);

    foreach ($groups as $group) {
      $members_list = $this->configuration['members'][$group->getGroupType()->id()] ?? NULL;
      if (!$members_list) {
        continue;
      }

      $group_relationship_type_ids = $grt_storage->getRelationshipTypeId($group->getGroupType()->id(), 'group_membership');

      // Get all members in the group.
      $query = $this->database->select('group_relationship_field_data', 'grfd');
      $query->fields('grfd', ['entity_id']);
      $query->condition('grfd.type', $group_relationship_type_ids, 'IN');
      $query->condition('grfd.gid', $group->id());

      if ($members_list === 'all') {
        $this->userIds = array_merge($this->userIds, $query->execute()->fetchCol());
      }
      elseif ($members_list === 'filtered_by_roles'
        || !empty($this->configuration['group_roles'][$group->getGroupType()->id()])) {
        $individual_members_ids = [];
        $insider_members_ids = [];
        $group_type_id = $group->getGroupType()->id();
        $group_roles = $this->ExtractRoles($group->getGroupType());

        // If no exclusion, get all members who's having at least one of the
        // selected roles.
        $exclusion = $this->configuration['exclusion'][$group->getGroupType()->id()] ?? NULL;
        if (!$exclusion) {
          if (!empty($group_roles[$group_type_id][PermissionScopeInterface::INDIVIDUAL_ID])) {
            $individual_members_ids = $this->getIndividualMembers(clone $query, $group_roles[$group_type_id][PermissionScopeInterface::INDIVIDUAL_ID]);
          }
          if (!empty($group_roles[$group_type_id][PermissionScopeInterface::INSIDER_ID])) {
            $insider_members_ids = $this->getInsiderMembers(clone $query, $group_roles[$group_type_id][PermissionScopeInterface::INSIDER_ID]);
          }
          $this->userIds = array_merge($this->userIds, $individual_members_ids, $insider_members_ids);
        }
        else {
          // Soft exclusion:
          // Exclude members having the selected roles. However, ignore the
          // exclusion if the member has roles other than the selected.
          if ($exclusion === self::EXCLUSION_SOFT) {
            // @todo handle the global roles.
            if (!empty($group_roles[$group_type_id][PermissionScopeInterface::INDIVIDUAL_ID])) {
              $query->leftJoin('group_content__group_roles', 'gcgr', 'grfd.id = gcgr.entity_id');
              $or_group = $query->orConditionGroup()
                ->condition('gcgr.group_roles_target_id', $group_roles[$group_type_id][PermissionScopeInterface::INDIVIDUAL_ID], 'NOT IN')
                ->condition('gcgr.group_roles_target_id', NULL, 'IS NULL');
              $query->condition($or_group);
            }

            $this->userIds = array_merge($this->userIds, $query->execute()->fetchCol());
          }
          // Strict exclusion:
          // Exclude members having at least one of the selected roles.
          elseif ($exclusion === self::EXCLUSION_STRICT) {
            if (!empty($group_roles[$group_type_id][PermissionScopeInterface::INDIVIDUAL_ID])) {
              $individual_members_ids = $this->getIndividualMembers(clone $query, $group_roles[$group_type_id][PermissionScopeInterface::INDIVIDUAL_ID]);
            }
            if (!empty($group_roles[$group_type_id][PermissionScopeInterface::INSIDER_ID])) {
              $insider_members_ids = $this->getInsiderMembers(clone $query, $group_roles[$group_type_id][PermissionScopeInterface::INSIDER_ID]);
            }
            $members_ids = array_merge($individual_members_ids, $insider_members_ids);
            if (!empty($members_ids)) {
              $query->condition('grfd.entity_id', $members_ids, 'NOT IN');
              $this->userIds = array_merge($this->userIds, $query->execute()->fetchCol());
            }
          }
        }
      }
    }

    return !empty($this->userIds)
      ? array_values($this->query($message)->execute())
      : [];
  }

  /**
   * {@inheritdoc}
   */
  public function query(MessageInterface $message): QueryInterface {
    $query = parent::query($message);
    $query->condition('uid', $this->userIds, 'IN');
    return $query;
  }

  /**
   * {@inheritdoc}
   */
  public function isApplicable(MessageInterface $message): bool {
    if ($this->hasAssociatedEntityField($message)) {
      $groups = $this->getGroupByMessage($message);
      return !empty($groups);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
        'members' => [],
        'group_roles' => [],
        'exclusion' => [],
      ] + parent::defaultConfiguration();
  }

  /**
   * Get members by individual roles.
   *
   * @param \Drupal\Core\Database\Query\SelectInterface $query
   *   The initial query.
   *
   * @return array
   *   List of members having the individual roles.
   */
  protected function getIndividualMembers(SelectInterface $query, array $individual_roles): array {
    $query->join('group_content__group_roles', 'gcgr', 'grfd.id = gcgr.entity_id');
    $query->condition('gcgr.group_roles_target_id', $individual_roles, 'IN');
    return $query->execute()->fetchCol();
  }

  /**
   * Get members by insider roles.
   *
   * @param \Drupal\Core\Database\Query\SelectInterface $query
   *   The initial query.
   *
   * @return array
   *   List of members having the insider roles.
   */
  protected function getInsiderMembers(SelectInterface $query, array $insider_roles): array {
    $query->join('user__roles', 'ur', 'grfd.entity_id = ur.entity_id');
    $query->condition('ur.roles_target_id', $insider_roles, 'IN');
    return $query->execute()->fetchCol();
  }

  /**
   * Get supported roles
   *
   * @param \Drupal\group\Entity\GroupTypeInterface $group_type
   *   The group type.
   *
   * @return array
   *   The supported roles.
   */
  protected function supportedRoles(GroupTypeInterface $group_type): array {
    $roles = [];
    foreach ($group_type->getRoles() as $group_role) {
      // Authenticated role is excluded because it is not a real role.
      // If we have it, it should return all members of the group. This is
      // already supported by another collector.
      if ($group_role->getScope() == PermissionScopeInterface::INDIVIDUAL_ID
        || ($group_role->getScope() === PermissionScopeInterface::INSIDER_ID
          && $group_role->getGlobalRole()->id() !== AccountInterface::AUTHENTICATED_ROLE)) {
        $roles[$group_role->id()] = $group_role->label();
      }
    }
    return $roles;
  }

  /**
   * Extract roles
   *
   * @param \Drupal\group\Entity\GroupTypeInterface $group_type
   *   The group type.
   *
   * @return array
   *   The extracted roles.
   */
  protected function ExtractRoles(GroupTypeInterface $group_type): array {
    $group_roles = [];
    foreach ($group_type->getRoles() as $role) {
      if (in_array($role->id(), $this->configuration['group_roles'][$group_type->id()])) {
        $group_roles[$group_type->id()][$role->getScope()][] = $role->getScope() === PermissionScopeInterface::INDIVIDUAL_ID
          ? $role->id()
          : $role->getGlobalRole()->id();
      }
    }

    return $group_roles;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state, MessageTemplateInterface $message_template): array {
    $form['description'] = [
      '#type' => 'item',
      '#title' => $this->t('Set filter settings per group type for this recipients collector:'),
      '#description' => $this->t('The settings will only be applied if the group from the associated entity matched with the group type.'),
    ];

    /** @var \Drupal\group\Entity\GroupTypeInterface[] $group_types */
    $group_types = $this->entityTypeManager->getStorage('group_type')->loadMultiple();

    foreach ($group_types as $group_type) {
      $form['settings'][$group_type->id()] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Members form %group_type', ['%group_type' => $group_type->label()]),
        '#tree' => TRUE,
      ];
      $form['settings'][$group_type->id()]['members'] = [
        '#type' => 'radios',
        '#options' => [
          'none' => $this->t('None: Does not retrieve any member'),
          'all' => $this->t('All members'),
          'filtered_by_roles' => $this->t('Filter members by roles'),
        ],
        '#default_value' => $this->configuration['members'][$group_type->id()] ?? [],
        '#description' => $this->t('Select list of members. Defaults to None.'),
      ];
      $form['settings'][$group_type->id()]['group_roles'] = [
        '#type' => 'checkboxes',
        '#title' => $this->t('Filter by roles:'),
        '#default_value' => $this->configuration['group_roles'][$group_type->id()] ?? [],
        '#options' => $this->supportedRoles($group_type),
        '#description' => $this->t('Select roles, to retrieve members having at least one of the selected roles. <br >In case of exclusion, members having the selected roles will be excluded.'),
        '#states' => [
          'visible' => [
            ':input[name="data[settings][' . $group_type->id() . '][members]"]' => [
              'value' => 'filtered_by_roles',
            ],
          ],
          'required' => [
            ':input[name="data[settings][' . $group_type->id() . '][members]"]' => [
              'value' => 'filtered_by_roles',
            ],
          ],
        ],
      ];

      // Add container for exclusion because radio doesn't work with states.
      $form['settings'][$group_type->id()]['exclusion_container'] = [
        '#type' => 'container',
        '#states' => [
          'visible' => [
            ':input[name^="data[settings][' . $group_type->id() . '][group_roles]["]' => [
              'checked' => TRUE,
            ],
            ':input[name="data[settings][' . $group_type->id() . '][members]"]' => [
              'value' => 'filtered_by_roles',
            ],
          ],
        ],
      ];
      $form['settings'][$group_type->id()]['exclusion_container']['exclusion'] = [
        '#type' => 'radios',
        '#title' => $this->t('Exclude members having the selected roles:'),
        '#options' => [
          'none' => $this->t('None'),
          self::EXCLUSION_SOFT => $this->t('Soft exclusion'),
          self::EXCLUSION_STRICT => $this->t('Strict exclusion'),
        ],
        '#default_value' => $this->configuration['exclusion'][$group_type->id()] ?? NULL,
      ];
      $form['settings'][$group_type->id()]['exclusion_container']['exclusion'][self::EXCLUSION_SOFT]['#description'] = $this->t('Exclude members having the selected roles. However, ignore the exclusion if the member has roles other than the selected. <br >WARNING: Not recommended because it does not yet support insiders.');
      $form['settings'][$group_type->id()]['exclusion_container']['exclusion'][self::EXCLUSION_STRICT]['#description'] = $this->t('Exclude members having at least one of the selected roles.');
    }
    return $form + parent::buildConfigurationForm($form, $form_state, $message_template);
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state, MessageTemplateInterface $message_template): void {
    parent::validateConfigurationForm($form, $form_state, $message_template);
    $values = $form_state->getValue('settings');
    $roles = $this->prepareRoles($form_state->getValue('settings'));
    $members_list = [];
    foreach ($values as $group_type_id => $value) {
      if ($value['members'] !== 'none') {
        $members_list[] = $value['members'];
      }
      if ($value['members'] === 'filtered_by_roles' && empty($roles[$group_type_id])) {
        $form_state->setErrorByName('settings][' . $group_type_id . '][group_roles', $this->t('You should select at least one role.'));
      }
    }
    if (empty(array_filter($members_list))) {
      $form_state->setErrorByName('', $this->t('You should select at least one list.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state, MessageTemplateInterface $message_template): void {
    parent::submitConfigurationForm($form, $form_state, $message_template);
    $this->configuration['group_roles'] = $this->prepareRoles($form_state->getValue('settings'));

    $exclusion = [];
    $members = [];
    foreach ($form_state->getValue('settings') as $group_type => $setting) {
      $members[$group_type] = $setting['members'];
      if ($setting['members'] === 'none') {
        $members[$group_type] = [];
      }
      if ($setting['members'] !== 'filtered_by_roles') {
        $exclusion[$group_type] = [];
      }
      elseif (!empty(array_filter($setting['group_roles'])) && $setting['exclusion_container']['exclusion']) {
        if ($setting['exclusion_container']['exclusion'] === 'none') {
          $setting['exclusion_container']['exclusion'] = [];
        }
        $exclusion[$group_type] = $setting['exclusion_container']['exclusion'];
      }
    }
    $this->configuration['members'] = array_filter($members);
    $this->configuration['exclusion'] = array_filter($exclusion);
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary(): array {
    $summary = [
      '#markup' => $this->description(),
    ];
    $list_summary = [];
    foreach ($this->configuration['members'] as $group_type => $members_list) {
      /** @var \Drupal\group\Entity\GroupTypeInterface $group_type */
      $group_type = $this->entityTypeManager->getStorage('group_type')->load($group_type);
      if ($members_list === 'all') {
        $list_summary[] = '<br/><strong>' . $group_type->label() . ':</strong> ' . $this->t('All members');
      }
      elseif ($members_list === 'filtered_by_roles') {
        $role_names = [];
        foreach ($this->configuration['group_roles'][$group_type->id()] as $group_role) {
          /** @var \Drupal\group\Entity\GroupRoleInterface $group_role */
          $group_role = $this->entityTypeManager->getStorage('group_role')->load($group_role);
          $role_names[] = $group_role->label();
        }
        $exclusion = '';
        if (!empty($this->configuration['exclusion'][$group_type->id()])) {
          $exclusion = $this->t('@exclusion exclude members having the following roles: ', [
            '@exclusion' => mb_strtoupper($this->configuration['exclusion'][$group_type->id()]),
          ]);
        }
        $list_summary[] = '<br/><strong>' . $group_type->label() . ':</strong> ' . $exclusion . ' ' . implode(', ', $role_names);
      }
    }
    $summary[] = [
      '#markup' => implode('', $list_summary),
    ];

    return [
        'recipient_collector' => $summary,
      ] + parent::getSummary();
  }

  /**
   * Prepare roles for saving.
   *
   * @param array $settings
   *   The settings grouped by group types.
   *
   * @return array
   *   List of group roles machine names grouped by group types.
   */
  protected function prepareRoles(array $settings): array {
    $roles = [];
    foreach ($settings as $group_type => $setting) {
      $filtered_roles = array_values(array_filter($setting['group_roles']));
      if ($setting['members'] !== 'filtered_by_roles') {
        $roles[$group_type] = [];
      }
      else if (!empty($filtered_roles)) {
        $roles[$group_type] = $filtered_roles;
      }
    }
    return array_filter($roles);
  }

  /**
   * Get groups by message's associated entity.
   *
   * @param \Drupal\message\MessageInterface $message
   *   The message entity.
   *
   * @return \Drupal\group\Entity\GroupInterface[]
   *   List group entities if found.
   */
  protected function getGroupByMessage(MessageInterface $message): array {
    $associated_entity = $this->getAssociatedEntity($message);
    if ($associated_entity instanceof GroupInterface) {
      return [$associated_entity];
    }
    elseif ($associated_entity instanceof GroupRelationshipInterface) {
      return [$this->getAssociatedEntity($message)->getGroup()];
    }
    elseif ($group_relationships = GroupRelationship::loadByEntity($associated_entity)) {
      $groups = [];
      foreach ($group_relationships as $group_relationship) {
        $groups[] = $group_relationship->getGroup();
      }

      return $groups;
    }
    return [];
  }

}
